/*! Copyright 2016-2021 Ling Zhang

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. */

declare class AgentFrameworkError extends Error {
}

interface Arguments {
    [index: number]: any;
    length: number;
}

interface Class<T extends object> extends Function {
    readonly prototype: T;
}

interface Design {
    readonly kind: number;
    readonly name: string;
    readonly declaringType: Function;
    readonly key: string | symbol;
    readonly type: Function | undefined;
}

interface Invocation<T extends Design = Design> {
    readonly design: T;
    invoke<T>(params: Arguments, receiver: unknown): T;
}

interface Interceptor<T extends Design = Design> {
    intercept(target: Invocation<T>, params: Arguments, receiver: unknown): unknown;
}

interface Attribute extends Object {
    beforeDecorate?(target: object | Function, key?: string | symbol, descriptor?: PropertyDescriptor | number): boolean;
    readonly interceptor?: Interceptor;
}

interface Filter<T> {
    (value: T, filterCriteria?: any): boolean;
}

interface MemberInfo extends Design {
    hasOwnAttribute<A1 extends Attribute>(type?: Class<A1>): boolean;
    getOwnAttribute<A2 extends Attribute>(type: Class<A2>): A2 | undefined;
    getOwnAttributes<A3 extends Attribute>(type?: Class<A3>): ReadonlyArray<A3>;
    addAttribute<A4 extends Attribute>(attribute: A4): void;
    findOwnAttributes<A5 extends Attribute>(filter: Filter<Attribute>, filterCriteria?: any): ReadonlyArray<A5>;
    hasOwnInterceptor(): boolean;
    getOwnInterceptors(): ReadonlyArray<Attribute>;
}

interface ParameterInfo extends MemberInfo {
    readonly index: number;
}

interface PropertyInfo extends MemberInfo {
    readonly descriptor: PropertyDescriptor | undefined;
    parameter(index: number): ParameterInfo;
    getParameter(index: number): ParameterInfo | undefined;
    getParameterTypes(): ReadonlyArray<any> | undefined;
    getParameters(): ReadonlyArray<ParameterInfo>;
    hasInterceptor(): boolean;
}

interface TypeInfo extends PropertyInfo {
    readonly static: TypeInfo;
    readonly prototype: TypeInfo;
    readonly type: Function;
    readonly base: TypeInfo | null | undefined;
    property(key: PropertyKey): PropertyInfo;
    hasOwnProperties(): boolean;
    getOwnProperties(): ReadonlyArray<PropertyInfo>;
    getOwnProperty(key: PropertyKey): PropertyInfo | undefined;
    getProperty(key: PropertyKey): PropertyInfo | undefined;
    findOwnProperties(filter: Filter<PropertyInfo>, filterCriteria?: any): ReadonlyArray<PropertyInfo>;
    findProperties(filter: Filter<PropertyInfo>, filterCriteria?: any): ReadonlyMap<TypeInfo, ReadonlyArray<PropertyInfo>>;
    findTypes(filter?: Filter<TypeInfo>, filterCriteria?: any): ReadonlyArray<TypeInfo>;
}

interface TypeInvocation extends Invocation<TypeInfo> {
}
interface PropertyInvocation extends Invocation<PropertyInfo> {
}
interface ParameterInvocation extends Invocation<ParameterInfo> {
}

interface ClassInterceptor extends Interceptor<TypeInfo> {
}
interface PropertyInterceptor extends Interceptor<PropertyInfo> {
}
interface ParameterInterceptor extends Interceptor<ParameterInfo> {
}

interface ClassAttribute extends Attribute {
    beforeDecorate?(target: Function): boolean;
    readonly interceptor?: ClassInterceptor;
}
interface PropertyAttribute extends Attribute {
    beforeDecorate?(target: object, key: string | symbol): boolean;
    readonly interceptor?: PropertyInterceptor;
}
interface ParameterAttribute extends Attribute {
    beforeDecorate?(target: Function, key: undefined, index: number): boolean;
    readonly interceptor?: ParameterInterceptor;
}

declare class MemberKinds {
    static None: number;
    static Class: number;
    static Static: number;
    static Property: number;
    static Parameter: number;
    static All: number;
}

declare function decorate<T extends Attribute>(attribute: T, allows?: number): (target: object | Function, targetKey?: string | symbol, descriptorOrIndex?: PropertyDescriptor | number) => void;

declare function decorateClass<T extends Attribute>(attribute: T): (target: Function) => void;

declare function decorateAgent<T extends Attribute>(attribute: T): (target: Function) => void;

declare function decorateMember<T extends Attribute>(attribute: T): (target: object | Function, targetKey: string | symbol, descriptor?: PropertyDescriptor | undefined) => void;

declare function decorateParameter<T extends ParameterAttribute>(attribute: T): (target: object | Function, targetKey: string | symbol | undefined, parameterIndex: number) => void;

declare function Reflector(target: Function | object): TypeInfo;

declare class Agent {
}

declare function agent(): ClassDecorator;

declare function singleton<T extends Function>(type?: T): PropertyDecorator;

declare function transit<T extends Function>(type?: T): PropertyDecorator;

declare function CreateAgent<T extends Function>(type: T, strategy?: ClassAttribute): T;

declare function IsAgent(target: Function | object): boolean;
declare function GetAgentType<T extends Function>(agent: T): T | undefined;
declare function GetType<T extends Function | object>(type: T): T | undefined;

declare function SetCustomInterceptor(type: Function, custom: Function, meta?: unknown): void;
declare function GetCustomInterceptor(type: Function): [Function, unknown?] | undefined;
declare function RemoveCustomInterceptor(type: Function): void;
declare function HasInterceptor(attribute: Attribute): boolean;
declare function GetInterceptor(attribute: Attribute): Interceptor | undefined;

declare function initializable(): ClassDecorator;

declare type InitializerHandler = () => void;
declare type StaticInitializerHandler = (target: TypeInvocation, params: Arguments, receiver: Function) => object;
declare const Initializer: unique symbol;

declare class AgentAttribute implements ClassAttribute, ClassInterceptor {
    get interceptor(): ClassInterceptor;
    intercept(target: TypeInvocation, params: any, receiver: Function): Function;
    construct<T extends Function>(this: any, target: T, params: Arguments, receiver: T): any;
}

declare function remember(key?: string): MethodDecorator;

declare function exclusive(uniqueId?: string): ClassDecorator;

export { Agent, AgentAttribute, AgentFrameworkError, Arguments, Attribute, Class, ClassAttribute, ClassInterceptor, CreateAgent, Design, Filter, GetAgentType, GetCustomInterceptor, GetInterceptor, GetType, HasInterceptor, Initializer, InitializerHandler, Interceptor, Invocation, IsAgent, MemberInfo, MemberKinds, ParameterAttribute, ParameterInfo, ParameterInterceptor, ParameterInvocation, PropertyAttribute, PropertyInfo, PropertyInterceptor, PropertyInvocation, Reflector, RemoveCustomInterceptor, SetCustomInterceptor, StaticInitializerHandler, TypeInfo, TypeInvocation, agent, decorate, decorateAgent, decorateClass, decorateMember, decorateParameter, exclusive, initializable, remember, singleton, transit };
